from collections import defaultdict
from multiprocessing import Pool
import os
import pickle
import pandas as pd
import time
import argparse


def GetConnectedComponents(df, all_ids=[], id1_col="id1", id2_col="id2_list", singles=[], silence=False):
    """
    expect df to be a dataframe with two columns
    the first column is the id of a sequence, and the second column is a list of ids of sequences that it is connected to (excluding self)
    singles is a list of all ids
    """
    sets = defaultdict(set)
    singles = []
    tested = {}
    n = 0
    ids = df[id1_col].values
    members = df[id2_col].values
    for i in range(len(ids)):
        id = ids[i]
        mems = members[i]
        if len(mems) == 1 and id in mems:
            singles.append(id)
            continue

        cur_set = set(list(mems) + [id])
        for m in cur_set:
            if m not in tested:
                tested[m] = True
        found = []
        for key, s in sets.items():
            if cur_set & s:
                found.append(key)
        if len(found) == 0:
            sets[n] = cur_set
            n += 1
            continue
        join_key = min(found)
        for key in [k for k in found if k != join_key]:
            sets[join_key] |= sets.pop(key, None)
        sets[join_key] |= cur_set
    if not silence:
        print(f"{len(sets)} clusters with {sum([len(x) for x in sets.values()])} sequences")
    added = 0
    for id in singles:
        # map the single clusters
        if id not in tested:
            sets[n] = set([id])
            tested[id] = True
            n += 1
            added += 1
            continue
    if not silence:
        print(f"{added} single nodes added")
    not_connected = [x for x in all_ids if x not in tested]
    for id in not_connected:
        sets[n] = set([id])
        n += 1
    if not silence:
        print(f"{len(not_connected)} unconnected nodes added")
    return sets

### Multiprocessing version
def ClusterChunkWrapper(args):
    return ClusterChunk(*args)

def ClusterChunk(chunk, chunk_id, cutoffs, id1_col, id2_col, score_col, out_dir, filter=None):
    if filter is not None:
        chunk = chunk[(chunk[id1_col].isin(filter)) & (chunk[id2_col].isin(filter))]
    
    for cutoff in cutoffs:
        file_dir=f"{out_dir}/{cutoff}"
        # make sure the directory exists
        if not os.path.exists(file_dir):
            os.makedirs(file_dir, exist_ok=True)

        # print(f"---cutoff: {cutoff}")
        chunk = chunk[(chunk[score_col] >= cutoff) | (chunk[id1_col] == chunk[id2_col])]

        df = chunk.groupby([id1_col])[id2_col].agg(set).reset_index()
        clusters = GetConnectedComponents(df, id1_col=id1_col, id2_col=id2_col, silence=True)
        pickle.dump(clusters, open(f"{file_dir}/{chunk_id}.pkl", "wb"))
        print(f"Chunk {chunk_id} cutoff {cutoff} done")

def ChunkInput(iterator, *args):
    for index, chunk in enumerate(iterator):
        # make list or tuple with index, chunk and rest of args
        yield chunk, index, *args

def ClusterStreamMulti(
    edge_table,
    cutoffs,
    query=0,
    target=1,
    score=11,
    label="cluster",
    directory=".",
    chunksize=10_000_000,
    has_header=False,
    field_separator="\t",
    filter_path=None,
    n_jobs=None):
    '''Multiprocessing version of cluster_stream
        Only handles the per-chunk clustering, not the final organization of clusters into a single file.

        Requires cluster_merging to be run after this function
    '''
    if has_header:
        header = 0
        for chunk in pd.read_csv(
            edge_table, delimiter=field_separator, chunksize=10, header=header
        ):
            qcol = chunk.columns[query]
            tcol = chunk.columns[target]
            scol = chunk.columns[score]
            break
    else:
        header = None
        qcol = query
        tcol = target
        scol = score

    # make sure the directory exists
    if not os.path.exists(directory):
        os.makedirs(directory, exist_ok=True)

    df_iterator = pd.read_csv(
        edge_table, delimiter=field_separator, chunksize=chunksize, header=header, usecols=[qcol, tcol, scol]
    )

    filter_set = set(pd.read_csv(filter_path, header=None)[0]) if filter_path else None

    with Pool(n_jobs) as pool:

        pool.imap_unordered(ClusterChunkWrapper, ChunkInput(df_iterator, cutoffs, qcol, tcol, scol, directory, filter_set))
        pool.close()
        pool.join()

### Old version
def Cluster(
    df_iterator, cutoffs, id1_col, id2_col, score_col, label, outdir, filter=None
):
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    start = time.time()
    i = 0
    sets = defaultdict(dict)
    c = 0
    cutoffs = sorted(cutoffs)
    for chunk in df_iterator:
        i += len(chunk)
        if filter is not None:
            chunk = chunk[(chunk[id1_col].isin(filter)) & (chunk[id2_col].isin(filter))]

        for cutoff in cutoffs:
            print(f"---cutoff: {cutoff}")
            chunk = chunk[(chunk[score_col] >= cutoff) | (chunk[id1_col] == chunk[id2_col])]

            df = chunk.groupby([id1_col])[id2_col].agg(set).reset_index()
            clusters = GetConnectedComponents(df, id1_col=id1_col, id2_col=id2_col)
            for cluster in clusters.values():
                found = []
                for cid, s in sets[cutoff].items():
                    if set.intersection(cluster, s):
                        found.append(cid)
                if not found:
                    sets[cutoff][c] = cluster
                    c += 1
                else:
                    min_id = min(found)
                    for cid in found:
                        cluster |= sets[cutoff].pop(cid, None)
                    sets[cutoff][min_id] = cluster

        print(i, f"\n{time.time()-start:.2f}s\n")

    # organize all cut-off data into dataframes
    clusters={}
    for cutoff in cutoffs:
        i=0
        data={}
        for _, s in sets[cutoff].items():
            for x in s:
                data[x]=i
            i += 1    
        clusters[cutoff]=pd.DataFrame.from_dict(data, orient='index').reset_index().rename(columns={'index':'id', 0:cutoff}) 

    # save all clusters in same file
    file=f"{label}_clusters.csv"
    file_path=os.path.join(outdir, file)
    table=None
    if os.path.exists(file_path):
        table=pd.read_csv(file_path)
        print(table.columns)
        table=table.drop(columns=[str(x) for x in cutoffs], errors='ignore')

    for cutoff in cutoffs:
        if table is None:
            table=clusters[cutoff]
        else:
            table=table.merge(clusters[cutoff], on='id', how='outer')
    table.to_csv(file_path, index=False) 

    return {'clusters':file}

def cluster_stream(
    edge_table,
    cutoffs,
    query=0,
    target=1,
    score=11,
    label="cluster",
    directory=".",
    chunksize=10_000_000,
    has_header=False,
    field_separator="\t",
    filter_path=None,
):
    '''Main cluster function for non-multiprocessing version'''
    if has_header:
        header = 0
        for chunk in pd.read_csv(
            edge_table, delimiter=field_separator, chunksize=10, header=header
        ):
            qcol = chunk.columns[query]
            tcol = chunk.columns[target]
            scol = chunk.columns[score]
            break
    else:
        header = None
        qcol = query
        tcol = target
        scol = score

    df_iterator = pd.read_csv(
        edge_table, delimiter=field_separator, chunksize=chunksize, header=header
    )
    filter_set = set(pd.read_csv(filter_path, header=None)[0]) if filter_path else None

    return Cluster(df_iterator, cutoffs, qcol, tcol, scol, label, directory, filter_set)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("-e", "-edge_table", help="Path to edge file", required=True)
    parser.add_argument(
        "-c",
        "-cutoffs",
        type=float,
        nargs="+",
        help="List of cutoffs for analysis.",
        required=True,
    )
    parser.add_argument(
        "-q",
        "-query",
        type=int,
        default=0,
        help="Index of column in edge table for QUERY id.",
    )
    parser.add_argument(
        "-t",
        "-target",
        type=int,
        default=1,
        help="Index of column in edge table for TARGET id.",
    )
    parser.add_argument(
        "-s",
        "-score",
        type=int,
        default=11,
        help="Index of column in edge table for SCORE.",
    )
    parser.add_argument(
        "-l", "-label", default="cluster", help="Outlabel, suffixes will be added."
    )
    parser.add_argument("-d", "-dir", default=".", help="Output directory")
    parser.add_argument(
        "-n",
        "-chunksize",
        type=int,
        default=10_000_000,
        help="Number of lines to read at once, adjust for memory consumption.",
    )
    parser.add_argument(
        "-H", "-Header", action="store_true", help="Flag for if input file has header"
    )
    parser.add_argument(
        "-f", "-field_separator", default="\t", help="Separator for table"
    )
    parser.add_argument("-filter", help="List of IDs to filter for")
    args = parser.parse_args()

    cluster_stream(
        args.e,
        args.c,
        args.q,
        args.t,
        args.s,
        args.l,
        args.d,
        args.n,
        args.H,
        args.f,
        args.filter,
    )



import pandas as pd
import numpy as np

def ClusterVectorRepresentative(clusters, level, vector_dict, target_clusters):
    cluster_ids=clusters.groupby(level)['id'].apply(list).to_dict()

    cluster_reps={}
    for c, ids in cluster_ids.items():
        if c not in target_clusters:
            continue
        members=np.array([vector_dict[x] for x in ids])
        ids=np.array(ids)
        representative=VectorRepresentative(members, ids)
        cluster_reps[c]=representative
        
    rep_df=pd.DataFrame(cluster_reps, index=[0]).T.reset_index().rename(columns={'index':'cluster', 0:'top_hit'})
    return rep_df

def VectorRepresentative(members, ids):
    mean=np.mean(members, axis=0)
    dist=np.linalg.norm(members-mean, axis=1)
    idx=np.argmin(dist)
    return ids[idx]


import argparse
import subprocess
import os

def RunMMseqs(fasta, out_dir='', prefix=''):
    # Run mmseqs
    outfile=prefix+'_mmseqs_search.tsv'
    mmseqs_args = ['mmseqs', 'easy-search', fasta, fasta, os.path.join(out_dir,outfile), 'tmp']
    opt_args=['--max-seqs', '1000', '--format-mode', '4', '--format-output', 'query,target,bits']

    mmseqs_args.extend(opt_args)
    subprocess.run(mmseqs_args)
    return {"mmseqs_result":outfile, "mmseqs_command":' '.join(mmseqs_args)}

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('fasta', type=str, required=True)
    parser.add_argument('out_label', type=str, required=True, help='Output label for the mmseqs search table')
    args = parser.parse_args()

    # Run mmseqs
    RunMMseqs(args.fasta, args.out_label)


from collections import defaultdict
import re
import subprocess
import tempfile
import os
from Bio import SeqIO
# import datetime
from zipfile import ZipFile, ZIP_DEFLATED
import argparse
import pandas as pd

# single letter list of amino acids, sorted by type
aa_list = ['H', 'K', 'R',                # (+)
           'D', 'E',                     # (-)
           'C', 'M', 'N', 'Q', 'S', 'T', # Polar-neutral
           'A', 'G', 'I', 'L', 'P', 'V', # Non-polar
           'F', 'W', 'Y',                # Aromatic
           '*']
_non_aa_re = re.compile(f'[^{"".join(aa_list)}]+')

def ReadFasta(file, id_index=None):
    seqs={}
    for record in SeqIO.parse(file, "fasta"):
        id=record.id
        if id_index is not None:
            id=id.split('|')[id_index]
        if _non_aa_re.search(str(record.seq)):
            print(f'Warning: {id} contains non-amino acid characters. Excluded from analysis.')
            continue
        seqs[id]=str(record.seq)
    return seqs

def FilterFastas(seqs, filter, id_map=None):
    out={}
    for k,v in seqs.items():
        if k in filter:
            id=id_map[k] if id_map is not None else k
            out[id]=v
    return out

def WriteFasta(file, seqs):
    with open(file,'w') as f:
        for k,v in seqs.items():
            
            f.write(f'>{k}\n{v}\n')

from multiprocessing import Pool
import shutil

def process_cluster(cluster_data):
    cluster, ids, fa, id_map, tmpdir = cluster_data

    fa_path = os.path.join(tmpdir, f'cluster_{cluster}.fa')
    msa_path = os.path.join(tmpdir, f'cluster_{cluster}.aln')
    hmm_path = os.path.join(tmpdir, f'cluster_{cluster}.hmm')
    hits_path = os.path.join(tmpdir, f'cluster_{cluster}.hits')

    seqs = FilterFastas(fa, ids, id_map)
    if seqs == {}:
        print(f'Warning: Cluster {cluster} is empty. Skipping.')
        return None
    WriteFasta(fa_path, seqs)
    
    # Run external commands
    subprocess.call(['mafft', fa_path], stdout=open(msa_path, 'w'), stderr=subprocess.DEVNULL)
    subprocess.call(['hmmbuild', hmm_path, msa_path], stdout=subprocess.DEVNULL)
    subprocess.call(['hmmsearch', '--tblout', hits_path, hmm_path, fa_path], stdout=subprocess.DEVNULL)

    # Process hits to find the top hit
    top_hit = None
    with open(hits_path) as hits:
        for line in hits:
            if not line.startswith('#'):
                top_hit = line.split()[0]
                break
    
    return cluster, msa_path, hmm_path, hits_path, top_hit

def ParseClusterLevel(cluster_file, level, fasta_file, target_clusters, zip_path, id_index=None, num_processes=None, id_map=None):
    '''
    cluster_file: csv file with columns 'id' and potentially multiple 'levels'
    level: the level to use for clustering
    fasta_file: fasta file with sequences, expected to have the same headers as the 'id' column in cluster_file
                - if using the cleaned fasta, ids will be in integers
    target_clusters: list of clusters to analyze
    zip_path: path to save the zip file
    id_index: index of the id in the fasta header assuming splitting on '|'
    num_processes: number of processes to use for parallelization, default uses all
    id_map: A filename, or dataframe (with 'id' and 'header' columns, where 'id' is the current label, and 'header' is the desired mapping) for changing the output headers from the original, e.g. from indices to original headers
    '''

    fa = ReadFasta(fasta_file, id_index)

    if type(cluster_file)==str:
        df=pd.read_csv(cluster_file, dtype={'id':str})
    else:
        df=cluster_file

    # collect id and level columns and convert to dictionary, group by level
    clusters=df.groupby(level)['id'].apply(list).to_dict()
    # print(clusters)

    if id_map is not None:
        if type(id_map)==str:
            id_map=pd.read_csv(id_map, dtype=str).set_index('id')['header'].to_dict()
        else:
            id_map=id_map.set_index('id')['header'].to_dict()

    # Prepare data for multiprocessing
    cluster_data_list = [(cluster, ids, fa, id_map, tempfile.mkdtemp()) for cluster, ids in clusters.items() if cluster in target_clusters]

    # Run analysis on each cluster in parallel
    with Pool(processes=num_processes) as pool:
        results = pool.map(process_cluster, cluster_data_list)

    # make sure output directory exists
    os.makedirs(os.path.dirname(zip_path), exist_ok=True)

    # Write results to zip
    with ZipFile(zip_path, 'w', compression=ZIP_DEFLATED) as zip_file:
        top = {}
        for result in results:
            if result is not None:
                cluster, msa_path, hmm_path, hits_path, top_hit = result
                top[cluster] = top_hit
                zip_file.write(msa_path, f'{cluster}/cluster_{cluster}.aln')
                zip_file.write(hmm_path, f'{cluster}/cluster_{cluster}.hmm')
                zip_file.write(hits_path, f'{cluster}/cluster_{cluster}.hits')
                # Clean up temporary files
                os.remove(msa_path)
                os.remove(hmm_path)
                os.remove(hits_path)

        # Add summary files to the zip
        top_table_str = 'cluster,top_hit\n' + '\n'.join([f'{k},{v}' for k, v in top.items()])
        zip_file.writestr('cluster_top_hits.csv', top_table_str)
        top_ids=[x for x in top.values() if x is not None]

        if id_map is not None:
            look_up=dict(zip(id_map.values(),id_map.keys()))
            top_ids=[look_up[x] for x in top_ids]
            # print(top_ids)
        zip_file.writestr('cluster_top_hits.fasta', ''.join([f'>{k}\n{v}\n' for k, v in FilterFastas(fa, top_ids, id_map).items()]))

    # Clean up temporary directories
    for _, _, _, _, tmpdir in cluster_data_list:
        shutil.rmtree(tmpdir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('cluster_file')
    parser.add_argument('fasta_file')
    parser.add_argument('--fasta_id_index', '-fi', type=int, help='Index of ID in fasta header')
    parser.add_argument('target_clusters')
    parser.add_argument('zip_path')
    args=parser.parse_args()

    target_clusters=set()
    with open(args.target_clusters) as f:
        for line in f:
            cluster=line.strip()
            if cluster=='':
                continue
            if cluster not in target_clusters:
                target_clusters.add(cluster)

    ParseClusterLevel(args.cluster_file, args.fasta_file, target_clusters, args.zip_path, args.fasta_id_index)


from dataclasses import dataclass
import time
from typing import Optional
import math
import Box2D
import matplotlib.pyplot as plt
import random
from tqdm import tqdm
import pandas as pd
import numpy as np

# Set up the Box2D world
# world = Box2D.b2World(gravity=(0,0))

@dataclass
class Circle():
    x:float
    y:float
    r:float
    id:str
    parent:Optional['Circle']=None
    size:Optional[float]=None

# Define a function to create a circle body
def create_circle(world, x, y, radius, density=1.0):
    body_def = Box2D.b2BodyDef()
    body_def.type = Box2D.b2_dynamicBody
    body_def.position = (x , y )
    body = world.CreateBody(body_def)

    shape = Box2D.b2CircleShape(radius=radius)
    fixture_def = Box2D.b2FixtureDef(shape=shape, density=density)
    body.CreateFixture(fixture_def)

    return body

def GetRadii(sizes, min_size=0, scale=1, force=None):
    # expect a series of cluster sizes
    if force is not None:
        sel=sizes[(sizes>=min_size)|(force)]
    else:
        sel=sizes[sizes>=min_size]
    radii=[]
    for s in sel:
        radii.append((s*scale/math.pi)**0.5)
    radii=pd.Series(radii, index=sel.index, name='radius')
    return radii

def MakeLayout(radii, radius_pad=1, space_mult=1.2, cycles=10_000, time_step=1, velocity_iterations=6, position_iterations=2,
                parent:Circle=None, seed=None, pull=0, min_radius_frac=.05, seed_coords=None, expand_mult=1, metadata=None, silence=True):
    '''
    radii: pandas Series of radii (index indicates the id of the circle)
    parent: parent circle to nest current circles within
    seed: random seed
    seed_coords: Series of (x,y) coordinates to use as starting points for circles, indexed by the id which is expected to match radii
    '''
    if seed is not None:
        random.seed(seed)

    # if parent is None:
    area=sum(math.pi*(r+radius_pad)**2 for r in radii)
    spacing=area**0.5 * space_mult

    world = Box2D.b2World(gravity=(0,0))
    # bodies=[]
    rmin, rmax = min(radii), max(radii)
    min_effective_r = rmax * min_radius_frac
    total_area=0
    start_circles=[]
    N=len(radii)
    lastx, lasty, lastr=0, 0, 0
    distances=[]
    radii_sum=[]
    for i, r in radii.items():
        if seed_coords is not None:
            row=seed_coords.loc[i] # match by index
            x, y=row['x'], row['y']
        else:
            angle=random.random() * 2 * math.pi
            pos_r=2**-math.log(max(min_effective_r, r)) * spacing + random.random() * spacing * .01
            x=pos_r*math.cos(angle)
            y=pos_r*math.sin(angle)
        total_area+=math.pi*(r+radius_pad)**2
        # bodies.append(create_circle(world, x, y, r+radius_pad, r**2))
        start_circles.append(Circle(x, y, r+radius_pad, i))
        # track distances
        if i != radii.index[0]:
            distances.append(((x-lastx)**2+(y-lasty)**2)**0.5)
            radii_sum.append(r+lastr)
        lastx, lasty, lastr=x, y, r
    if len(start_circles)>1:
        avg_dist=sum(distances)/len(distances)
        avg_radii_overlap=sum([radii_sum[i]-distances[i] for i in range(len(radii_sum))])/len(radii_sum)
    else:
        avg_dist=0
        avg_radii_overlap=0

    # get boundaries
    xmax, xmin, ymax, ymin=GetLimits(start_circles)
    # space the circles out so that the bounding box is 1.2 times the area of the circles
    if seed_coords is None:
        expand_factor=1
        pull_mod=1
    else:
        expand_factor=(total_area/(xmax-xmin)/(ymax-ymin)) * expand_mult * len(radii)**.5 * avg_radii_overlap
        pull_mod=expand_factor/cycles
    
    # get centroid of xy
    # if expand_mult!=1:
    cx, cy=0, 0
    for c in start_circles:
        cx+=c.x
        cy+=c.y
    cx/=len(start_circles)
    cy/=len(start_circles)

    start=time.time()
    if N >100 and not silence:
        print(f' Average distance: {avg_dist}, average overlap: {avg_radii_overlap}')
        print(f' Expanding by {expand_factor} for {len(radii)} circles')

    # scale distance from centroid
    for c in start_circles:
        c.x=(c.x-cx)*expand_factor+cx
        c.y=(c.y-cy)*expand_factor+cy

    if N >100 and not silence:
        print(f'  Expanding took {time.time()-start} seconds')

    bodies=[]
    for c in start_circles:
        bodies.append(create_circle(world, c.x, c.y, c.r, c.r**2))

    for cycle in range(cycles):
        cycle_mod=2 if cycle<cycles*2/3 else 1
        if N>1000:
            cycle_start=time.time()
        if pull!=0:
            for c in bodies:
                x,y=c.position
                mod=c.fixtures[0].density * pull * pull_mod
                c.ApplyForceToCenter((-x*mod,-y*mod), True)
        world.Step(time_step*cycle_mod, velocity_iterations, position_iterations)
        if N>1000 and not silence:
            print(f'  Cycle took {time.time()-cycle_start} seconds')

    if N>100 and not silence:
        print(f' Layout took {time.time()-start} seconds')
        

    circles=[] 
    for i in range(len(bodies)):
        x,y=bodies[i].position
        id=radii.index[i]
        circles.append(Circle(x, y, radii[id], id, parent))
    xmax, xmin, ymax, ymin=GetLimits(circles)

    # get center to off set grouping
    cx=(xmax+xmin)/2    
    cy=(ymax+ymin)/2

    xoff, yoff= 0, 0
    if parent is not None:
        xoff, yoff=parent.x, parent.y
    for c in circles:
        c.x=c.x-cx+xoff
        c.y=c.y-cy+yoff

    return circles, metadata

def GetLimits(circles, pad=0):
    xmax, xmin, ymax, ymin=circles[0].x, circles[0].x, circles[0].y, circles[0].y
    for c in circles:
        xmax=max(xmax, c.x+c.r)
        xmin=min(xmin, c.x-c.r)
        ymax=max(ymax, c.y+c.r)
        ymin=min(ymin, c.y-c.r)
    return xmin-pad, xmax+pad, ymin-pad, ymax+pad

if __name__=='__main__':

    # test code
    n=10_000
    radii=GetRadii(pd.Series([random.random()*100+1 for x in range(n)]))
    circles=MakeLayout(radii)
    xmin, xmax, ymin, ymax=GetLimits(circles)

    f, ax=plt.subplots(figsize=(10,10))
    ax.axis('off')

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    for c in circles:
        ax.add_patch(plt.Circle((c.x, c.y), c.r, alpha=.2, linewidth=1, fill=False))
    plt.show()




import pickle
import os
from multiprocessing import Pool
import random

def merge_overlapping_sets(list1, list2):
    """
    Input requires the sets in each list to be disjoint amongst each othor, but not between lists.
    """
    merged_sets = []  # To keep track of merged sets for later addition
    to_remove = set()  # To keep track of indices in list2 that have been merged

    # filter out singles
    singles = [s for s in list1 if len(s) == 1]
    singles.extend([s for s in list2 if len(s) == 1])
    list1 = [s for s in list1 if len(s) > 1]
    list2 = [s for s in list2 if len(s) > 1]

    disjoint=[]
    # Iterate over each set in the first list
    for set1 in list1:
        # Iterate over each set in the second list by index and value
        overlapping=[]
        for i, set2 in enumerate(list2):
            if i in to_remove:
                # Skip sets that are already merged
                continue
            # Check for intersection
            if not set1.isdisjoint(set2):
                overlapping.append(i)

        # Add the merged set to the list of merged sets
        if len(overlapping) > 0:
            list2[overlapping[0]].update(set1, *[list2[i] for i in overlapping[1:]])
            for i in overlapping[1:]:
                list2[i]=None
                to_remove.add(i)  # Mark this set as merged
        else:
            disjoint.append(set1)

    # Rebuild list2 without the merged sets
    list2[:] = [s for i, s in enumerate(list2) if i not in to_remove]
    list2.extend(disjoint)

    grouped=set()
    for i, s in enumerate(singles):
        (item,)=s
        for set2 in list2:
            if item in set2:
                grouped.add(i)
                break
    singles = [s for i, s in enumerate(singles) if i not in grouped]
    list2.extend(singles)
    
    return list2

def merge_pair(pair):
    """
    Wrapper function for merging a pair of datasets.
    """
    data=[]
    for item in pair:
        if type(item) is str:
            # load the file and take the dict values
            with open(item, 'rb') as f:
                item = list(pickle.load(f).values())
        data.append(item)

    return merge_overlapping_sets(data[0], data[1])

def pairwise_merge(datasets, max_datasets=100):
    """
    Function to merge datasets pairwise using multiprocessing.
    """
    if len(datasets) < 2:
        return datasets
    # shuffle the datasets
    random.shuffle(datasets)

    # If we have more datasets than the maximum, hold out the extras
    hold_out=[]
    if len(datasets) > max_datasets:
        hold_out = datasets[max_datasets:]
        datasets = datasets[:max_datasets]

    # Ensure we have an even number of datasets for pairing; if not, strip the last element
    left_over = None
    if len(datasets) % 2 != 0:
        left_over = datasets.pop()

    # Pair up datasets for merging
    pairs = zip(*[iter(datasets)]*2)

    # Use multiprocessing Pool to merge pairs in parallel
    with Pool() as pool:
        merged_datasets = list(pool.imap_unordered(merge_pair, pairs)) # using list forces blocking

    if hold_out:
        merged_datasets.extend(hold_out)
    if left_over:
        # If we stripped a dataset, add it back to the merged datasets
        merged_datasets.append(left_over)

    return merged_datasets

import time
def iterative_merge(datasets, max_datasets=None):
    """
    Repeatedly apply pairwise merging until one dataset remains.
    """
    cycle=1
    while len(datasets) > 1:
        if max_datasets is None:
            max_datasets = len(datasets)
        print('Cycle:',cycle, f'  Datasets: {min(len(datasets), max_datasets)}/{len(datasets)}')
        start=time.time()
        datasets = pairwise_merge(datasets, max_datasets) # actual function
        print(f' - completed in {time.time()-start:.2f} s')
        cycle+=1
    return datasets[0]

import pandas as pd
def SetsToDataFrame(dataset):
    """
    Convert a list of sets to a DataFrame.
    """
    
    # sort by size
    dataset = sorted(dataset, key=len, reverse=True)

    data={}
    for i, s in enumerate(dataset):
        for x in s:
            data[x]=i
    return pd.DataFrame.from_dict(data, orient='index').reset_index().rename(columns={'index':'id', 0:'cluster'})


# test bokeh
from bokeh.plotting import figure, show, output_notebook
from bokeh.models import WheelZoomTool, BoxZoomTool, HoverTool, ColumnDataSource, CustomJS, LegendItem, Legend, ColorBar, LinearColorMapper
from bokeh.models.widgets import Select
from bokeh.layouts import column, row
from ..layout.circle_collision import Circle
from collections import defaultdict
import re


def CoordinatesToCircles(xs, ys, ids):
    '''
    Convert a list of x and y coordinates to a list of Circle objects.

    Arguments:
    - xs: list
        list of x coordinates
    - ys: list
        list of y coordinates
    - ids: list
        list of ids

    Returns:
    - circles: list
        list of Circle objects
    '''
    return [Circle(x,y,1,id) for x,y,id in zip(xs,ys,ids)]

rgba_re=re.compile(r'rgba\((\d+),(\d+),(\d+),(\d+)\)')
output_notebook()
black='rgba(0,0,0,255)'
def CirclePlot(layout, 
               size=800,
               annot_text:list=None, annot_top_n:int=None, annot_na=True, singleton_annot_text=False, # annotation related
               annot_colors=None, base_fill=None, # face color related
               outlines=None, base_line='rgba(0,0,0,255)', base_line_width=1, highlight_line_width=1, # outline related
               fixed_point_size=None # if set, overrides individual radii
               ):
    '''
    Make a hierarchical circle plot using bokeh. Adds colors and annotations to the circles.

    Arguments:
    - General:
        - layout: dict
            layout of the circles
        - size: int
            size of the plot
    - Tooltip text:
        - annot_text: list
            list of annotation dictionaries generated with AnnotCluster
        - annot_top_n: int
            number of highest proportion entries to show in the text annotations
        - annot_na: bool
            whether to show NA values in the text annotations
        - singleton_annot_text: bool
            assumes data is singletons, to allow faster mapping of annotations to data points
    - Fill coloring:
        - annot_colors: dict
            dictionary of annotation colors, generated with ColorAnnot
        - base_fill: str
            default fill color for the circles if there are no annotations
    - Outline coloring:
        - outlines: dict
            dictionary of outline colors, generated with ColorAnnot
        - base_line: str
            default outline color for the circles if there are no annotations
        - base_line_width: int
            default outline width for the circles if there are no annotations
        - highlight_line_width: int
            width of the outline for highlighted circles
    - Other:
        - fixed_point_size: int
            if set, overrides individual radii

    Returns:
    - p: bokeh.plotting.figure
        bokeh figure object
    - manual_plot_data: dict
        data used for plotting, in a format that can be used to replicate the plot in matplotlib
    '''
    wheel_zoom=WheelZoomTool()
    wheel_zoom.zoom_on_axis=False
    box_zoom=BoxZoomTool()
    box_zoom.match_aspect=True

    # make figure
    fig_args={'width':size, 'height':size, 'match_aspect':True, 'min_border':50,
                'tools':['pan', wheel_zoom, box_zoom, 'reset', 'save']}
    p=figure(**fig_args)
    p.toolbar.active_scroll=wheel_zoom

    # plotting circles
    renderers={}
    levels=list(layout.keys())
    manual_plot_data=defaultdict(list)
    annot_hover=[]
    for tgt_level in levels:
        x=[c.x for c in layout[tgt_level].values()]
        y=[c.y for c in layout[tgt_level].values()]
        r=[c.r for c in layout[tgt_level].values()]
        colors=[annot_colors[tgt_level][c.id] if tgt_level in annot_colors and c.id in annot_colors[tgt_level] else base_fill for c in layout[tgt_level].values()] if annot_colors is not None else [base_fill]*len(x)
        line_colors=[outlines[tgt_level][c.id] if tgt_level in outlines and c.id in outlines[tgt_level] else base_line for c in layout[tgt_level].values()] if outlines is not None else [base_line]*len(x)
        line_widths=[highlight_line_width if tgt_level in outlines and c.id in outlines[tgt_level] else base_line_width for c in layout[tgt_level].values()] if outlines is not None else [base_line_width]*len(x)

        draw_dict={'x':x, 'y':y, 'r':r, 'colors':colors, 'line_colors':line_colors, 'line_widths':line_widths,
                            'name':[c.id for c in layout[tgt_level].values()], 
                            'level':[tgt_level]*len(x),
                            'size':[c.size for c in layout[tgt_level].values()]
                            }
        # add text
        if annot_text is not None:
            for annot in annot_text:
                annot_label=f'{annot['value']} {annot['method']}'.capitalize()
                annot_key=f'{annot['value']}_{annot['method']}'
                annot_tooltip=(annot_label, f'@{annot_key}{{safe}}')
                if annot_tooltip not in annot_hover:
                    annot_hover.append(annot_tooltip)
                if tgt_level in annot:
                    if singleton_annot_text:
                        level_annot_dict=annot[tgt_level].set_index('id')['value'].to_dict()
                        hover_text=[level_annot_dict[c.id] if c.id in level_annot_dict else None for c in layout[tgt_level].values()]
                    else:
                        hover_text=MakeHoverText(annot, tgt_level, draw_dict['name'], top_n=annot_top_n, 
                                             blank_fill=None, cluster_sizes=draw_dict['size'], annot_na=annot_na)
                else:
                    hover_text=[None]*len(x)
                draw_dict[annot_key]=hover_text

        circles=ColumnDataSource(data=draw_dict)
        if fixed_point_size is None:
            renderers[tgt_level]=p.circle('x', 'y', radius='r', fill_color='colors', line_color='line_colors', source=circles, line_width='line_widths')
        else:
            renderers[tgt_level]=p.scatter('x', 'y', size=fixed_point_size, fill_color='colors', line_color='line_colors', source=circles, line_width='line_widths')

        # save data for manual plotting
        manual_plot_data['x'].extend(x)
        manual_plot_data['y'].extend(y)
        manual_plot_data['r'].extend(r if fixed_point_size is None else [fixed_point_size]*len(x))
        c_conv=[]
        lc_conv=[]
        for i in range(len(x)):
            if colors[i] is None:
                c_conv.append([1,1,1,0])
            else:
                c_conv.append([float(v)/255 for v in rgba_re.match(colors[i]).groups()])
            if line_colors[i] is None:
                lc_conv.append([1,1,1,0])
            else:
                lc_conv.append([float(v)/255 for v in rgba_re.match(line_colors[i]).groups()])
        manual_plot_data['colors'].extend(c_conv)
        manual_plot_data['line_colors'].extend(lc_conv)
        manual_plot_data['line_widths'].extend(line_widths)
        manual_plot_data['name'].extend([c.id for c in layout[tgt_level].values()])
        manual_plot_data['level'].extend([tgt_level]*len(x))
    
    # Define tooltips
    no_size=all([v is None for v in draw_dict['size']])
    tooltips = [
        ("Level", "@level"),
        ("Cluster", "@name"),
        # ("(x,y)", "($x, $y)"),
    ] + ([("Size", "@size")] if not no_size else []) + annot_hover

    hover=HoverTool(tooltips=tooltips, renderers=list(renderers.values())[-1:])
    p.add_tools(hover)

    # selection for which tooltips to show
    dropdown=Select(title='Label level', options=levels, value=levels[-1])
    # Callback function to update hover renderers
    callback = CustomJS(args=dict(hover=hover, options=renderers), code="""
        const value = cb_obj.value;
        const renderer = options[value];
        hover.renderers = [renderer];
    """)
    
    dropdown.js_on_change('value', callback)
    
    # Layout
    legends=[]
    if annot_colors is not None:
        legend, manual_data=MakeLegend(annot_colors, p)
        manual_plot_data['fill_legend']=manual_data
        legends.append(legend)
    if outlines is not None:
        legend, manual_data=MakeLegend(outlines, p, outline=True)
        manual_plot_data['line_legend']=manual_data
        legends.append(legend)

    layout=row(*([p] + legends + [dropdown]))
    show(layout)

    return p, manual_plot_data

def MakeLegend(annot_colors, p, outline=False):
    legend_fig = figure(width=200, height=p.height, toolbar_location=None, min_border=0, outline_line_color=None)
    legend_fig.xaxis.visible = False
    legend_fig.yaxis.visible = False
    legend_fig.xgrid.visible = False
    legend_fig.ygrid.visible = False

    manual_plot_data={}            
    if 'categories' in annot_colors:
        legend_items=[]
        manual_plot_data['legend_colors']={}
        for cat, color in annot_colors['categories'].items():
            # if outline:
            #     legend_items.append(LegendItem(label=cat, renderers=[legend_fig.circle(0,0, color=['grey'], line_color=[color], radius=0)]))
            # else:
            legend_items.append(LegendItem(label=cat, renderers=[legend_fig.circle(0,0, color=[color], line_color=['black'], radius=0)]))
            manual_plot_data['legend_colors'][cat]=[float(v)/255 for v in rgba_re.match(color).groups()]
        legend=Legend(items=legend_items)
        legend_fig.add_layout(legend)
        # modify legend text size
        legend_fig.legend.label_text_font_size = '9pt'
        legend_fig.legend.spacing = 1
    elif 'colorscale' in annot_colors:
        colorscale=[v[1] for v in annot_colors['colorscale']]
        for i, v in enumerate(colorscale):
            colorscale[i]=[float(x)/255 for x in rgba_re.match(v).groups()]
        manual_plot_data['colorscale']=colorscale
        manual_plot_data['colorscale_range']=[annot_colors['min'], annot_colors['max']]
        # convert to hex rgb
        colorscale=[f'#{"".join([hex(int(x*255))[2:].zfill(2) for x in c])}' for c in colorscale]
        color_mapper=LinearColorMapper(palette=colorscale, low=annot_colors['min'], high=annot_colors['max'])
        legend=ColorBar(color_mapper=color_mapper, label_standoff=12, location=(0,0))
        legend_fig.add_layout(legend, 'left')
        # plot 1 invisible circle to stop error message
        legend_fig.circle(0,0, color='white', line_color='white', radius=0)
    # add title to legend
    legend_fig.title.text_font_size='10pt'
    legend_fig.title.align='center'
    legend_fig.title.text=annot_colors['value']+' '+ annot_colors['method'] + (' (fill)' if not outline else ' (outline)')
    manual_plot_data['title']=legend_fig.title.text
    return legend_fig, manual_plot_data

def MakeHoverText(annot, level, ids, top_n=None, blank_fill=None, cluster_sizes=None, annot_na=True):
    hover_text=[] # one per id, in the order they appear
    is_numeric=annot['method']!='counts'

    for i, id in enumerate(ids):
        cluster_size=None if cluster_sizes is None else cluster_sizes[i]
        if id not in annot[level].id.values:
            hover_text.append(blank_fill)
        else:
            data=annot[level].loc[annot[level].id==id].copy()
            if not annot_na:
                data=data.dropna(subset=['value'])
            if is_numeric:
                hover_text.append(f'{data.value.iloc[0]:.2f}')
            else:
                total=data['count'].sum() if cluster_size is None else cluster_size # explicit will be more accurate as some may just be missing annotations
                data['frac']=(data['count']/total*100).round(2)
                item_total=len(data)
                if top_n is not None:
                    data=data.nlargest(top_n, 'count')
                row_text=[]
                for row in data.itertuples():
                    row_text.append(f'{row.value}: {row.count} ({row.frac}%)')
                if item_total!=len(data):
                    row_text.append(f'...{item_total-len(data)} more')
                hover_text.append('<br>'.join(row_text))
    return hover_text

#####
# Saving figures as PDF
#####

# saving plot with bokeh is finicky, use matplotlib instead
import matplotlib.pyplot as plt
import matplotlib as mpl
from ..layout.circle_collision import GetLimits

def ExportFigure(layout, plot_data, size=5, legend_padding=1, as_scatter=False, fixed_point_size=None):
    '''
    Export a plot from bokeh by replicating the same plot in matplotlib.

    Arguments:
    - layout: dict
        layout of the circles
    - plot_data: dict
        data used for plotting, output from CirclePlot
    - size: int
        Height of the plot in inches. Extra width will be added for legends
    - legend_padding: int
        Factor for extra space added for each legend. Default figure is 3, and each legend is 1.
    - as_scatter: bool
        If True, plot as scatter plot instead of circles (much faster with a lot more points). Circle plots cannot use scatter, as the radii do not scale properly.
    - fixed_point_size: int
        If set, overrides individual radii. Use for setting a specific size for scatters with uniform size.

    Returns:
    - f: matplotlib.figure.Figure
    '''
    all_c=[]
    for lv, c_dict in layout.items():
        all_c+=list(c_dict.values())
    xmin, xmax, ymin, ymax=GetLimits(all_c, 5)

    legends_expected=sum([1 for key in ['fill_legend', 'line_legend'] if key in plot_data])
    width_ratios=[3]+[legend_padding]*legends_expected
    size_extra=size * sum(width_ratios[1:])/width_ratios[0] if legends_expected>0 else 0
    f, axes = plt.subplots( 1, 1+legends_expected, figsize=(size+size_extra, size), gridspec_kw={'width_ratios': width_ratios})
    ax=axes[0]
    ax.axis('off')
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.set_aspect('equal', adjustable='box')
    # remove padding on margins
    if as_scatter:
        sizes=plot_data['r'] if fixed_point_size is None else fixed_point_size
        ax.scatter(plot_data['x'], plot_data['y'], s=sizes, c=plot_data['colors'], edgecolors=plot_data['line_colors'], linewidths=plot_data['line_widths'])
    else:
        for i in range(len(plot_data['x'])):
            circle=plt.Circle((plot_data['x'][i], plot_data['y'][i]), plot_data['r'][i], 
                            facecolor=plot_data['colors'][i], edgecolor=plot_data['line_colors'][i], lw=plot_data['line_widths'][i])
            ax.add_patch(circle)

    legend_count=1
    for key in ['fill_legend', 'line_legend']:
        if key not in plot_data:
            continue
        lax = axes[legend_count]
        lax.axis('off')
        legend_data=plot_data[key]
        # make legend if data given
        if 'legend_colors' in legend_data:
            legend_items=[]
            for cat, color in legend_data['legend_colors'].items():
                legend_items.append(plt.Line2D([0], [0], marker='o', linestyle='None', color=color, markeredgecolor='black', label=cat, markersize=5, markeredgewidth=1))
            # place legend outside to the right, also remove frame
            leg=lax.legend(handles=legend_items, loc='center', title=legend_data['title'], bbox_to_anchor=(0.5, 0.5), frameon=False)
            # Customize the legend title
            plt.setp(leg.get_title(), fontweight='bold')
        elif 'colorscale' in legend_data:
            # make color bar
            custom_cmap=mpl.colors.ListedColormap(legend_data['colorscale'])
            norm=mpl.colors.Normalize(vmin=legend_data['colorscale_range'][0], vmax=legend_data['colorscale_range'][1])
            scalar_mappable=mpl.cm.ScalarMappable(norm=norm, cmap=custom_cmap)
            plt.colorbar(scalar_mappable, ax=lax, label=legend_data['title'], fraction=.2, anchor=(0,0.5))
        legend_count+=1
    # f.subplots_adjust(wspace=0, hspace=0)
    f.tight_layout(w_pad=0)
    return f


import pandas as pd
from sklearn.cluster import MiniBatchKMeans, BisectingKMeans
import multiprocessing as mp
import numpy as np

def Kmeans(embeddings, n_clusters, outlabel):
    # convert embeddings dict to numpy ndarray
    ids=np.array(list(embeddings.keys()))
    X=np.array(list(embeddings.values()))

    # run kmeans, using minibatch kmeans for speed
    if type(n_clusters)==int:
        n_clusters=[n_clusters]

    clusters={}
    for k in n_clusters:
        kmeans = MiniBatchKMeans(n_clusters=k, random_state=0)
        kmeans.fit(X)
        clusters[k]=kmeans.labels_
    
    # make dataframe and write to file
    df=pd.DataFrame.from_dict(clusters)
    df['id']=ids
    df=df[['id']+n_clusters]
    out_name=f'{outlabel}_kmeans.csv'
    df=df.sort_values('id', key=lambda x: x.astype(int))
    df.to_csv(out_name, index=False)
    return {'kmeans': out_name}

def KMeansBisecting(embeddings, n_clusters, ids, out_dir='', out_prefix=''):

    # run kmeans, using minibatch kmeans for speed
    if type(n_clusters)==int:
        n_clusters=[n_clusters]

    # clusters={}
    for k in n_clusters:
        print(f'Running bisecting kmeans with {k} clusters')
        kmeans = BisectingKMeans(n_clusters=k, random_state=0)
        kmeans.fit(embeddings)
        # make a dataframe, with id and cluster assignments
        
        pd.DataFrame({'id':ids, 'cluster':kmeans.labels_}).to_csv(f'{out_dir}{out_prefix}_kmeans_bisecting_{k}.csv', index=False)



import os, shutil
import pickle
from .esm_extract import create_parser, run
import torch
from Bio import SeqIO

def Embed(fasta, out_dir='', prefix='', tmp_dir='tmp/esm/', tok_per_batch=30000, cont_run=False):

    # clear tmp_dir if it already exists
    if os.path.exists(tmp_dir) and not cont_run:
        shutil.rmtree(tmp_dir)

    if cont_run and os.path.exists(tmp_dir):
        # read all file names in tmp_dir
        files=os.listdir(tmp_dir)
        # get basename without path and extensions
        ids=set([f.split('.')[0] for f in files if f.endswith('.pt')])

        # read in fasta file, get ids of stuff yet to be processed
        records=SeqIO.parse(fasta, 'fasta')
        yet_to_process=[x for x in records if x.id not in ids]

        # write fasta to tmp_dir
        fasta_in=os.path.join(tmp_dir, 'tmp.fasta')
        with open(fasta_in, 'w') as f:
            SeqIO.write(yet_to_process, f, 'fasta')
        print(f'Continuing embedding remaining {len(yet_to_process)} sequences')
    else:
        fasta_in=fasta

    args_str=['esm1b_t33_650M_UR50S', fasta_in, tmp_dir, '--toks_per_batch' , str(tok_per_batch), '--include', "mean"]
    parser=create_parser()
    args=parser.parse_args(args_str)
    run(args)

    embed_file=f'{prefix}_embeddings.pkl'
    embeddings={}
    # open torch models and extract embeddings
    for f in os.listdir(tmp_dir):
        if not f.endswith('.pt'):
            continue
        model=torch.load(os.path.join(tmp_dir, f))
        id=f.split('.')[0]
        embeddings[id]=model['mean_representations'][33]
    # pickle embeddings
    with open(os.path.join(out_dir,embed_file), 'wb') as f:
        pickle.dump(embeddings, f)

    # remove tmp_dir
    shutil.rmtree(tmp_dir) 

    return {'embeddings': embed_file}


from Bio import SeqIO
import pandas as pd
import os

def sanitize_fasta_headers(fasta_file, out_dir='', out_prefix=''):
    
    cleaned_filename=out_prefix+'_cleaned.fasta'
    cleaned_file = open(os.path.join(out_dir,cleaned_filename), 'w')
    mapping_filename=out_prefix+'_header_map.txt'

    i=0
    mapping={}
    for record in SeqIO.parse(fasta_file, 'fasta'):
        id=record.id
        print(f'>{i}\n{record.seq}', file=cleaned_file)
        mapping[i]=id
        i+=1
        
    cleaned_file.close()
    mapping=pd.DataFrame.from_dict(mapping, orient='index').reset_index()
    mapping.columns=['id','header']
    mapping.to_csv(os.path.join(out_dir,mapping_filename), index=False)

    return {'cleaned_fasta': cleaned_filename, 'header_map': mapping_filename}